#ifndef __JABYENGINE_CD_INTERNAL_HPP__
#define __JABYENGINE_CD_INTERNAL_HPP__
#include <PSX/System/IOPorts/cd_io.hpp>
#include <PSX/AutoLBA/auto_lba.hpp>

namespace JabyEngine {
    namespace CD {
        namespace internal {
            enum struct State {
                Free = 0,
                Done = 0,

                Reading,
                BufferFull,
                Error,
            };

            struct DataSector {
                static constexpr size_t SizeBytes = 2048;
                static constexpr size_t SizeWords = (SizeBytes/sizeof(uint32_t));

                uint32_t data[SizeWords];

                template<typename T>
                static constexpr T words_to_sectors(T size) {
                    return (size + static_cast<T>(DataSector::SizeWords - 1))/static_cast<T>(DataSector::SizeWords);
                }
            };

            struct FileInfo {
                uint16_t lba;
                uint16_t sectors;

                static constexpr FileInfo from_auto_lba(const AutoLBAEntry& entry) {
                    return FileInfo{entry.lba, DataSector::words_to_sectors(entry.size_words)};
                }
            };

            class SectorBufferAllocator {
            private:
                typedef DataSector* (*AllocatorFunction)(void* ctx);

            private:
                void*             ctx      = nullptr;
                AllocatorFunction allocate = nullptr;

                constexpr SectorBufferAllocator(void* ctx, AllocatorFunction function) : ctx(ctx), allocate(function) {
                }

            public:
                static constexpr SectorBufferAllocator create(void* obj, AllocatorFunction function) {
                    return SectorBufferAllocator(obj, function);
                }
            };

            extern VolatilePOD<CD_IO::Interrupt::Type> last_interrupt;

            struct Command {
                static void wait_until(CD_IO::Interrupt::Type irq) {
                    while(last_interrupt.read() != irq);
                }

                template<typename...ARGS>
                static void send(CD_IO::CommandFifo_t& cmd_fifo, CD_IO::ParameterFifo_t& parameter_fifo, CD_IO::Command::Info cmd, ARGS...args) {
                    while(CD_IO::IndexStatus.read().is_bit_set(CD_IO::IndexStatus::IsTransmissionBusy));

                    (parameter_fifo.write(static_cast<uint8_t>(args)), ...);
                    cmd_fifo.write(cmd.id);
                }

                template<typename...ARGS>
                static void send_wait(CD_IO::CommandFifo_t& cmd_fifo, CD_IO::ParameterFifo_t& parameter_fifo, CD_IO::Command::Info cmd, ARGS...args) {
                    send(cmd_fifo, parameter_fifo, cmd, args...);
                    wait_until(cmd.complete_irq);
                }
            };

            State read_file(FileInfo file_info, const SectorBufferAllocator& buffer_allocator);
        }
    }
}
#endif //!__JABYENGINE_CD_INTERNAL_HPP__